/*
 * Copyright (C) 1995-2002 Neil Peter Charley.
 * Portions Copyright (C) 1996 James William Hawtin.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * dns.h
 *
 * Header file for DNS lookup server code.
 *
 */

/*
   $Revision: 1.2 $ 
 */

#if !defined(DNSH)
#define DNSH

#include <time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>

#if defined(__linux__)
#define SIGTYPE int it
#define USE_SIGACTION
#define USE_SIGEMPTYSET
#define HAVE_UNISTDH
#endif /* __linux__ */

/*
   Toggle debug mode 
 */
/*
   #define DEBUG_DNS 
 */

#define AWAITING_DNS		-1
#define DNS_SERVER_PID_PATH	"run/dns_server.pid"
#define DNS_LOOKUP_ATTEMPTS	3

/*
   The various fd's for the two pipes between client and server 
 */

#define DNS_SERVER_READ		dns_toserver_fds[0]
#define DNS_SERVER_WRITE	dns_toclient_fds[1]
#define DNS_CLIENT_READ		dns_toclient_fds[0]
#define DNS_CLIENT_WRITE	dns_toserver_fds[1]

/*
   And the various fd's for between the server and the lookup children 
 */

#define DNS_PARENT_READ		from_child_fds[0]
#define DNS_PARENT_WRITE	to_child_fds[1]
#define DNS_CHILD_READ		to_child_fds[0]
#define DNS_CHILD_WRITE		from_child_fds[1]

/*
   Messages between client and server 
 */

#define DNS_SERVER_CONNECT_MSG "Server Connect:"
#define DNS_CLIENT_CONNECT_MSG "Client Connect:"
#define DNS_CLIENT_SEND_REQUEST 'A'
#define DNS_CLIENT_CANCEL_REQUEST 'B'
#define DNS_SERVER_SEND_REPLY ':'

/*
   Misc #define's 
 */

#define MAX_DNS_IN_PROGRESS 5	/*
				   Unused atm 
				 */

/*
   Typedef's 
 */

typedef int              dns_identifier;

/*
   Struct for a dns request 
 */

struct s_dns_request
{
	dns_identifier           dns_id;
	struct in_addr           ip_addr;
	time_t                   request_time;
	int                      using;		 /*
						    dns_child slot number 
						  */
	struct s_dns_request    *next;
};
typedef struct s_dns_request dns_request;

struct s_dns_child
{
	int                      to_child_fds[2];
	int                      from_child_fds[2];
	pid_t                    pid;
	struct s_dns_request    *request;
};
typedef struct s_dns_child dns_child;

#endif /*
          DNSH 
        */
