/*
 * Copyright (C) 1995-2002 Neil Peter Charley.
 * Portions Copyright (C) 1996 James William Hawtin.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * dns_client.c
 *
 */

char                     dnsclientc_rcsid[] = "$Revision: 1.5 $";


#if defined(__STRICT_ANSI__)
 #include <features.h>
 #define __USE_POSIX
 #define __USE_BSD
 #define __USE_MISC 
#endif /* __STRICT_ANSI__ */

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#if defined(HAVE_UNISTDH) || !defined(LOTHLORIEN)
#include <unistd.h>
#endif /* HAVE_UNISTDH */
#include <signal.h>
#include <sys/ioctl.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/un.h>
#if defined(HAVE_FILIOH)
#include <sys/filio.h>
#endif /* HAVE_FILIOH */
#include <sys/wait.h>

#if defined(LOTHLORIEN)
#define NEED_FREE
#define NEED_STRDUP
#include "mem.h"
#include "misc.h"
#include "comms.h"
#else /* !LOTHLORIEN */
#include "player.h"
#include "mem.h"
#endif /* LOTHLORIEN */
#include "dns.h"
#if defined(USING_DMALLOC)
 #include <dmalloc.h>
#endif /* USING_DMALLOC */

#if defined(ANSI)
 extern int kill __P ((__pid_t __pid, int __sig));
 extern void bzero __P ((__ptr_t __s, size_t __n));
#endif
/* Extern functions */

extern char             *end_string(char *str);
extern void              log(char *,...);
extern void              setup_itimer(void);

/*
   Extern variables 
 */

#ifndef errno
extern int               errno;
#endif

#if defined(LOTHLORIEN)
extern connection       *firstc;

#define PCSTRUCT	 connection
#else /*
         !LOTHLORIEN 
       */
extern player           *flatlist_start;

#define PCSTRUCT	 player
#endif /*
          LOTHLORIEN 
        */

/*
   Local functions 
 */

int                      init_dns_server(void);
void                     kill_dns_server(void);
void                     dns_process_reply(int msg_size);
void                     read_dns_reply(void);
void                     send_dns_request(PCSTRUCT * c, struct in_addr *sadd);

/*
   Local Variables 
 */

#define DNS_BUFFER_SIZE (256 + sizeof(dns_identifier))	/*
							   See RFC822 
							 */
int                      dns_toclient_fds[2];
int                      dns_toserver_fds[2];
int                      dns_server_pid = 0;
dns_identifier           dns_id = 0;
char                     dns_buf_input[DNS_BUFFER_SIZE + 1];
char                     dns_buf_output[DNS_BUFFER_SIZE + 1];
char                     reply_buf[DNS_BUFFER_SIZE + 1];


/*
 * Start up the dns server
 */

int                      init_dns_server(void)
{
	int                      ret;

#if defined(FIONBIO)
	int                      dummy;

#endif /*
          FIONBIO 
        */
	fd_set                   fds;
	struct timeval           timeout;

	if (-1 == pipe(dns_toclient_fds))
	{
		switch (errno)
		{
		case EMFILE:
			log("boot", "init_dns_server: Too many fd's in use by"
			    " process");
			exit(1);
		case ENFILE:
			log("boot", "init_dns_server: Too many fd's in use in"
			    " system");
			exit(1);
		case EFAULT:
			log("boot", "init_dns_server: dns_toclient_fds invalid!");
			exit(1);
		}
	}
	if (-1 == pipe(dns_toserver_fds))
	{
		switch (errno)
		{
		case EMFILE:
			log("boot", "init_dns_server: Too many fd's in use by"
			    " process");
			exit(1);
		case ENFILE:
			log("boot", "init_dns_server: Too many fd's in use in"
			    " system");
			exit(1);
		case EFAULT:
			log("boot", "init_dns_server: dns_toserver_fds invalid!");
			exit(1);
		}
	}
	ret = fork();
	switch (ret)
	{
	case -1:	/*
			   Error 
			 */
		log("boot", "init_dns_server couldn't fork!");
		exit(1);
	case 0:	/*
			   Child 
			 */
		close(DNS_CLIENT_READ);
		close(DNS_CLIENT_WRITE);
		close(0);
		dup(DNS_SERVER_READ);
		close(DNS_SERVER_READ);
		close(1);
		dup(DNS_SERVER_WRITE);
		close(DNS_SERVER_WRITE);
		execlp("bin/dns_server", "dns_server", 0);
		log("boot", "init_dns_server failed to exec dns_server");
		exit(1);
	default:	/* Parent */
		dns_server_pid = ret;
		close(DNS_SERVER_READ);
		close(DNS_SERVER_WRITE);
		DNS_SERVER_READ = DNS_SERVER_WRITE = -1;
#if defined(FIONBIO)
		if (ioctl(DNS_CLIENT_READ, FIONBIO, &dummy) < 0)
		{
			log("error", "Ack! Can't set non-blocking to dns read.");
		}
		if (ioctl(DNS_CLIENT_WRITE, FIONBIO, &dummy) < 0)
		{
			log("error", "Ack! Can't set non-blocking to dns write.");
		}
#endif /* FIONBIO */
	}

	FD_ZERO(&fds);
	FD_SET(DNS_CLIENT_READ, &fds);
	timeout.tv_sec = 15;
	timeout.tv_usec = 0;
	while (-1 == (ret = select(FD_SETSIZE, &fds, 0, 0, &timeout)))
	{
		if (errno == EAGAIN || errno == EINTR)
		{
			continue;
		}
		log("boot", "init_dns_server: Timed out waiting for server"
		    " connect");
		if (ret == -1)
		{
			perror("DNS client");
		}
		kill_dns_server();
		return 0;
	}
	ioctl(DNS_CLIENT_READ, FIONREAD, &ret);
	while (ret != strlen(DNS_SERVER_CONNECT_MSG))
	{
		sleep(1);
		ioctl(DNS_CLIENT_READ, FIONREAD, &ret);
	}
	ret = read(DNS_CLIENT_READ, dns_buf_input, ret);
	dns_buf_input[ret] = '\0';
	if (strcmp(dns_buf_input, DNS_SERVER_CONNECT_MSG)
	    || ret != strlen(DNS_SERVER_CONNECT_MSG))
	{
		fprintf(stderr, "From DNS: '%s'\n", dns_buf_input);
		log("boot", "init_dns_server: Bad connect from server, killing");
		kill_dns_server();
		return 0;
	}
	log("boot", "DNS Server Up and Running");
	setup_itimer();

	return 1;
}

/*
   Shutdown the dns server 
 */

void                     kill_dns_server(void)
{
	close(DNS_CLIENT_READ);
	close(DNS_CLIENT_WRITE);
	DNS_CLIENT_READ = -1;
	DNS_CLIENT_WRITE = -1;
	/*
	   In case it has been suspended 
	 */
	kill(dns_server_pid, SIGCONT);
	kill(dns_server_pid, SIGTERM);
	sleep(3);
	kill(dns_server_pid, SIGKILL);
	waitpid(dns_server_pid, NULL, 0);
}

void                     send_dns_request(PCSTRUCT * c, struct in_addr *inadd)
{
	char                    *s;
	int                      bwritten;

	s = dns_buf_output;
	*s++ = DNS_CLIENT_SEND_REQUEST;
	memcpy(s, &dns_id, sizeof(dns_id));
	s += sizeof(dns_id);
	memcpy(s, &(inadd->s_addr), sizeof(inadd->s_addr));
	s += sizeof(inadd->s_addr);
	bwritten = write(DNS_CLIENT_WRITE, dns_buf_output,
			 (s - dns_buf_output));
	c->dns_id = dns_id++;
	if (bwritten < (s - dns_buf_output))
	{
		log("dns", "DNS Client failed to write request, killing and restarting Server");
		kill_dns_server();
		init_dns_server();
		bwritten = write(DNS_CLIENT_WRITE, dns_buf_output,
				 (s - dns_buf_output));
		if (bwritten < (s - dns_buf_output))
		{
			log("dns", "Restart failed");
		}
	}
#if defined(DEBUG_DNS_TOO)
	fprintf(stderr, "DNS Bytes Written %d, Should have sent %d\n",
		bwritten, (s - dns_buf_output));
	fprintf(stderr, "DNS Client: %08X\n",
		(int)ntohl(inadd->s_addr));
	fflush(stderr);
#endif /*
          DEBUG_DNS_TOO 
        */
}

void                     read_dns_reply(void)
{
	int                      bread;
	int                      toread;
	int                      i;
	static int               bufpos = 0;

	ioctl(DNS_CLIENT_READ, FIONREAD, &toread);
	if (toread <= 0)
	{
		return;
	}
	bread = read(DNS_CLIENT_READ, dns_buf_input, DNS_BUFFER_SIZE);
	dns_buf_input[bread] = '\0';

	for (i = 0; i < bread;)
	{
		reply_buf[bufpos++] = dns_buf_input[i++];
		if ((bufpos > (sizeof(char) + sizeof(dns_identifier)))
		  &&                       (reply_buf[bufpos - 1] == '\n'))
		{
			dns_process_reply(bufpos);
			bufpos = 0;
			ZERO(reply_buf, DNS_BUFFER_SIZE);
		}
	}
}

void                     dns_process_reply(int msg_size)
{
	char                    *s;
	int                      i;
	dns_identifier           id;
	PCSTRUCT                *c;

#if defined(LOTHLORIEN)
#define FORC(c) for (c = firstc ; c ; c = c->next)
#else /*
         !LOTHLORIEN 
       */
#define FORC(c) for (c = flatlist_start ; c ; c = c->flat_next)
#endif /*
          LOTHLORIEN 
        */

	for (i = 0; i < msg_size;)
	{
		switch (reply_buf[i++])
		{
		case DNS_SERVER_SEND_REPLY:
			memcpy(&id, &reply_buf[i], sizeof(dns_identifier));
			i += sizeof(dns_identifier);
			FORC(c)
			{
				if (c->dns_id == id)
				{
					break;
				}
			}
#if defined(DEBUG_DNS_TOO)
			fprintf(stderr, "DNS Client: Got reply '%s'\n", &reply_buf[i]);
#endif /*
          DEBUG_DNS_TOO 
        */
			/* Strip the '\n' on the end of the string as we don't want it in what we'll be printing */
			s = strchr(&reply_buf[i], '\n');
			if (s)
			{
				*s = '\0';
			}
			if (c)
			{
#if defined(LOTHLORIEN)
				FREE(c->hostname);
				c->hostname = STRDUP(&reply_buf[i]);
#else /* !LOTHLORIEN */
				strncpy(c->inet_addr, &reply_buf[i],
					MAX_INET_ADDR);
				c->inet_addr[MAX_INET_ADDR - 1] = '\0';
#endif /* LOTHLORIEN */
			} else
			{
				/* Can only assume connection dropped from
				 * here and we still somehow got a reply,
				 * throw it away 
				 */
			}
			/* We're done, so set i to fall out of the for() (can't break; because we're in an enclosed switch() ) */
			i = msg_size;
			break;
		default:
			log("dns", "Bad reply from server '%d'", reply_buf[i]);
			i++;
		}
	}
}
